"""
Module containing docker interface functionality.

This contains a `DockerClient` class that will be used to interface with the docker client on 
  the user's system.
Having a docker client installed on the system is a prerequisite for this to work.
"""

import docker  # type: ignore
from docker.models.images import Image  # type: ignore
from integrate_ai.utils.rest_client import RestClient
from integrate_ai.utils.error_handling import DockerException
from typing import Optional, Dict, List
from distutils.version import LooseVersion
from typing import Any


class DockerClient:
    """Class to interact with docker client.

    This class performs various docker operations such as login, pull, rmi, etc.
    """

    def __init__(self, token: Optional[str] = None, gpu: bool = False, image_name: str = "fl-client"):
        """Initialize.

        Args:
            token (str): IAI token.
        """
        self.token: Optional[str] = token
        self.gpu = gpu
        self.rest_client: Optional[RestClient] = None
        if self.token:
            self.rest_client = RestClient(self.token)
        self.docker_client: docker.DockerClient = docker.from_env()
        self.repo: str = ""
        self.ecr_token: str = ""
        self.ecr_username: str = ""
        self.image_name = image_name

    def login(self) -> Optional[Dict]:
        """Retrieves authentication to connect to ECR repo and performs docker login.

        Returns:
            login_response (Dict): Contains response information returned from docker login.

        Raises:
            DockerException: If error performing docker login command.
        """
        if self.rest_client:
            ecr_response = self.rest_client.get_ecr_token()
            self.ecr_token = ecr_response.get("authorizationToken", "")
            self.repo = ecr_response.get("proxyEndpoint", "")
            self.ecr_username = ecr_response.get("username", "")
            try:
                login_response = self.docker_client.login(
                    username=self.ecr_username,
                    password=self.ecr_token,
                    registry=self.repo,
                )
                return login_response
            except Exception as e:
                raise DockerException(e)
        else:
            raise DockerException("No REST Client available to connect ")

    def get_repo_name(self) -> str:
        """Retrieves the client image name to pull.

        Returns:
            name (str): name of the docker image to retrieve.
        """
        if self.rest_client:
            ecr_repo_response = self.rest_client.get_ecr_repo(image=self.image_name)
        return ecr_repo_response["repo_name"]

    def _auth_dict(self) -> Dict[str, str]:
        """Returns authorization dictionary for docker commands using ECR repo.

        Returns:
            auth_dict (Dict): contains authorization dictionary for ECR connection
        """
        return {"username": self.ecr_username, "password": self.ecr_token}

    def filter_image_by_tag_part(self, image: Image, tag_part: str) -> bool:
        """Determines if `tag_part` is found in image tags.

        Args:
            image (Image): image to check tags for.
            tag_part (str): Checked against image tags to see if is contained in first part of tag.

        Returns:
            tag_part_contained (bool): True if `tag_part` found in tags, False otherwise.
        """
        for tag in image.tags:
            if tag_part == tag.split(":")[0]:
                return True
        return False

    def get_local_versions(self, tag_part: str) -> List[Image]:
        """Returns the local images that contain `tag_part` as part of one of their tags.

        Args:
            tag_part (str): prefix to filter images by.

        Returns:
            image_list (List[Image]): list of images on user's system containing `tag_part`.
        """
        return [image for image in self.docker_client.images.list() if self.filter_image_by_tag_part(image, tag_part)]

    def get_latest_version_of_image(self, images: List[Image]) -> Optional[str]:
        """Returns the latest version number from a list of comparable images.

        Note: If 'latest' is the only version found (no numbered version) or list is empty, will return `None`

        Args:
            images (List[Image]): list of images to search for latest version.

        Returns:
            version_no (str) latest version number found if available. Otherwise `None`.
        """
        local_versions = []
        for image in images:
            for tag in image.tags:
                repo, version = tag.split(":")
                local_versions.append(version)
        non_latest_versions = [el for el in local_versions if not el.startswith("latest")]

        if len(non_latest_versions) > 0:
            return sorted(
                non_latest_versions,
                key=LooseVersion,
                reverse=True,
            )[0]
        else:
            return None

    def get_versions(self):
        """Returns a list of available versions from the ecr repository.

        Returns:
            versions (List[string]): list of available docker images to pull.
        """
        # Split the repository name into the registry and the short form of the repository
        versions_response = self.rest_client.get_ecr_versions(image=self.image_name)
        return versions_response["versions"]

    def delete_images(self, images: List[Image]) -> None:
        """Deletes images contained in a list

        Args:
            images (List[Image]): list of images to delete.
        """
        for image in images:
            self.docker_client.images.remove(image.id, force=True)

    def pull(self, repo: str, tag: str) -> Dict[str, str]:
        """Pulls docker image corresponding to specified repo and tag.

        Args:
            repo (str): image name to pull.
            tag (str): tag of image to pull.

        Returns:
            pull_response (Dict[str,str]): Response Dict returned from docker pull request.

        Raises:
            DockerException: If error in docker pull command.
        """
        try:
            pull_response = self.docker_client.images.pull(repository=repo, tag=tag, auth_config=self._auth_dict())
            return pull_response
        except Exception as e:
            raise DockerException(e)

    def get_latest_available_version(self) -> str:
        """Returns the latest available version of the client docker available in repo.

        Returns:
            latest_version (str): latest client image version number available in repo."""

        if self.rest_client:
            is_cpu = self.image_name == "fl-client" and not self.gpu
            version_response = self.rest_client.get_latest_ecr_version(image=self.image_name, is_cpu=is_cpu)
            return version_response["version"]
        else:
            raise DockerException("No REST Client available to connect ")

    def run(self, image: str, detach: bool, options: Dict[str, Any]) -> Optional[Dict]:
        """Starts a container with given image name and configurations.

        Args:
            image str: image name to run the container
            detach bool: whether or not the process should return immediately
            options Dict[str, str]: a dict of options accepted by `docker run` command.

        Returns:
            run_response (Dict[str,str]): Response Dict returned from docker run command.

        Raises:
            DockerException: If error in docker run command.

        """
        try:
            run_response = self.docker_client.containers.run(image, detach=detach, **options)

            return run_response
        except Exception as e:
            raise DockerException(e)

    def get_container(self, container_name):
        """Gets container with given name.

        Args:
            container_name str: name of container we are searching for

        Returns:
            Container: class representation of container object

        Raises:
            DockerException: If error in getting container.

        """
        try:
            return self.docker_client.containers.get(container_name)
        except Exception as e:
            raise DockerException(e)



"""Module containing main CLI entrypoint. Performs any preprocessing required before handing off to Typer."""

import os
import sys


def interactive_mode_check():
    """Check if --no-prompt given as kwarg and set env variable accordingly."""
    if "--no-prompt" in sys.argv:
        os.environ["IAI_DISABLE_PROMPTS"] = "true"
        sys.argv.remove("--no-prompt")
    else:
        os.environ["IAI_DISABLE_PROMPTS"] = ""


def main():
    """Run any pre-Typer checks before starting Typer app."""
    interactive_mode_check()
    from integrate_ai.main import main as cli_main

    cli_main()


if __name__ == "__main__":
    main()  # pragma: no cover



from pathlib import Path
import os

user_root_dir = Path.home()
logs_folder = os.path.join(user_root_dir, "iai", "logs")


def get_log_paths(client_name: str, session: str):
    """Folder paths for logs can be used when given a client name and session id

    Args:
        client_name (str): name of the client
        session (str): session id

    Returns:
        session_log_path (str): path to the session folder
        full_log_path (str): full path to the client log file
    """
    session_log_path = os.path.join(logs_folder, session)
    full_log_path = os.path.join(session_log_path, f"{client_name}.log")

    return session_log_path, full_log_path



"""
Module containing REST interface functionality.

This contains a `RestClient` class that will be used to interface with the IAI REST API.
"""

import jwt
from typing import Any, Dict
from integrate_ai.utils.error_handling import check_for_IntegrateAIException
import requests
import os


class RestClient:
    """Client class for interacting with the integrate.ai API.

    This class includes functionality for performing various operations using
      the integrate.ai REST API.
    This class requires a valid token before it can perform its operations.
    """

    def __init__(self, token: str):
        """Initiates a client given a valid decoded IAI token object.

        Args:
            token (str): IAI token."""

        self.auth_token: str = token
        decoded_token: Dict[str, str] = self._decode_jwt_token(token=token)
        self.customer: str = decoded_token["customer"]
        self.env: str = decoded_token["env"]
        self.api_url: str = self._get_api_endpoint()
        self.gateway_url: str = self._get_gateway_endpoint()

    def _headers(self) -> Dict[str, str]:
        """Add authenticating headers for use with REST API."""

        header = {"content-type": "application/json"}
        header["Authorization"] = f"Bearer {self.auth_token}"
        return header

    def _decode_jwt_token(self, token: str, secret=None, verify_signature: bool = False) -> Dict[str, str]:
        """Decodes a JWT token."""
        decoded_token = jwt.decode(
            token,
            secret,
            algorithms=["HS256"],
            options={"verify_signature": verify_signature},
        )
        return decoded_token

    def _get_api_endpoint(self) -> str:
        """Returns API endpoint."""

        return os.environ.get("IAI_DEBUG_API_URL") or f"https://iai-api-{self.customer}-{self.env}.integrateai.net"

    def _get_gateway_endpoint(self) -> str:
        """Returns Gateway endpoint."""

        return (
            os.environ.get("IAI_DEBUG_GATEWAY_URL")
            or f"https://iai-fl-gateway-{self.customer}-{self.env}.integrateai.net"
        )

    def make_get_request(self, url: str, body={}, headers=None) -> requests.Response:
        """Performs a GET request using client.

        Args:
            url (str): URL to make GET request to.
            body (Dict): body to include in GET request.
            headers (Dict): The headers to be added to the GET request.
                Defaults to the integrate.ai API Authorization headers.

        Returns:
            response (requests.Response): response object from GET request.

        Raises:
            IntegrateAIException: If error in GET request."""
        request_headers = self._headers()
        if headers:
            request_headers.update(headers)
        response = requests.get(url, headers=request_headers, json=body)
        check_for_IntegrateAIException(response)
        return response

    def get_ecr_token(self) -> Dict[str, str]:
        """Retrieves an ECR token using IAI token.

        Returns:
            ecr_response (Dict[str,str]): response to ECR GET request."""
        url = f"{self.api_url}/fl/ecr/token"
        ecr_response = self.make_get_request(url).json()

        return ecr_response

    def get_ecr_repo(self, image: str) -> Dict[str, str]:
        """Retrieves the ecr repo name for a package"""
        url = f"{self.api_url}/fl/ecr/{image}/repo"
        return self.make_get_request(url).json()

    def get_ecr_versions(self, image: str) -> Dict[str, str]:
        """Retrieves the versions of an image from the ecr repo"""
        url = f"{self.api_url}/fl/ecr/{image}/list"
        return self.make_get_request(url).json()

    def get_latest_ecr_version(self, image: str, is_cpu: bool) -> Dict[str, str]:
        """Retrieves the latest version of an image in ecr"""
        url = f"{self.api_url}/fl/ecr/{image}/latest"
        if is_cpu:
            url += "?cpu"
        return self.make_get_request(url).json()

    def get_pip_install_command(self, version="", package="") -> Dict[str, str]:
        """Get pip install command to install sdk.
        Args:
            version (str): Version of sdk to install.
            package (str): Package name to install.
        Returns:
                A Dict of the form
                ```
                {
                    "pipInstallCommand":"<pip_install_command>",
                }
                ```
        Raises:
            IntegrateAIException: Customized IntegrateAI exception for the HTTP Exception.
        """
        url = f"{self.api_url}/fl/pip/{package}"
        if version:
            url += f"/{version}"
        response = requests.get(
            url,
            headers=self._headers(),
        )
        check_for_IntegrateAIException(response=response)

        return response.json()

    def get_package_versions(self, package="") -> Dict[str, Any]:
        """Get package versions.
        Args:
            package (str): Name of package that has a list of available versions.
        Returns:
                A Dict of the form
                ```
            {
                "package_name": "<package-name>",
                "package_versions": [
                    {
                        "version": <version_number_1>,
                        "revision": <revision_string>,
                        "status": "Published"
                    },
                    {
                        "version": <version_number_2>,
                        "revision": <revision_string>,
                        "status": "Published"
                    }
                ]
            }
                ```
        Raises:
            IntegrateAIException: Customized IntegrateAI exception for the HTTP Exception.
        """
        url = f"{self.api_url}/fl/pip/{package}/list"
        response = requests.get(
            url,
            headers=self._headers(),
        )
        check_for_IntegrateAIException(response=response)

        return response.json()



"""
Module containing miscellaneous typer utilities.
"""

from functools import wraps
from pathlib import Path
from typing import Callable, Any, Dict
import typer
import os
import re
import subprocess
from integrate_ai.utils.error_handling import IntegrateAIException


def toggle_prompt_wrapper(option_func: Callable, disable_prompts: str) -> Callable:
    """Decorates a function to remove prompt as a kwarg if disable_prompts is 'true'.

    Args:
        option_func (Callable): function to wrap.
        disable_promtps (str): whether to disable prompts or not ('true' if prompts should be disabled).

    Returns:
        decorated function (Callable): either same function as option_func or same function with prompt kwarg removed.
    """

    @wraps(option_func)
    def wrapper_func(*args, **kwargs):
        updated_kwargs = kwargs.copy()
        if disable_prompts == "true":
            updated_kwargs = kwargs.copy()
            updated_kwargs.pop("prompt")
            return option_func(*args, **updated_kwargs)
        else:
            return option_func(*args, **updated_kwargs)

    return wrapper_func


# wrapper around typer.Option to enable prompting to be condition on IAI_DISABLE_PROMPTS env variable
TogglePromptOption = toggle_prompt_wrapper(typer.Option, disable_prompts=os.environ.get("IAI_DISABLE_PROMPTS", ""))


def cast_path_to_dict(path: Path) -> Dict[str, str]:
    return {
        "parent_path": str(path.parent),
        "parent_dir": str(path.parts[-2]) or "",
        "full_path": str(path),
    }


def path_param_callback(path: str) -> Dict[str, Any]:
    # Expect URLs to be properly formed (s3://).
    # Otherwise they will be treated like local path.
    if path.lower().startswith("s3://"):
        return {
            "s3": True,
            "url": path,
        }
    else:
        if path.lower().startswith("~"):
            p = Path(path).expanduser()
        else:
            p = Path(path).resolve()
        validate_posix_path(p)
        return cast_path_to_dict(p)


def validate_posix_path(path: Path) -> None:
    if not (path.is_file() or path.is_dir()):
        raise typer.BadParameter(f"file or directory {str(path)} does not exist.")
    if not os.access(str(path), os.R_OK):
        raise typer.BadParameter(f"file or directory {str(path)} is not readable.")


def show_package_version(package):
    show_version_command = f"pip show {package}"
    print(f"Getting current version for {package}.")

    pipe = subprocess.Popen(show_version_command, stdout=subprocess.PIPE, shell=True)
    out, err = pipe.communicate()
    if pipe.returncode != 0:
        raise IntegrateAIException("Failed to get version.", err)
    out_string = out.decode("utf-8")
    for line in out_string.splitlines():
        if re.search("Version", line):
            print(line)
            break
